# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Models for the workflows db. Based on schema v9
"""

from sqlalchemy import (
    Boolean,
    Column,
    DateTime,
    ForeignKey,
    Index,
    Integer,
    String,
    Text,
    UniqueConstraint,
    func,
)
from sqlalchemy.orm import declarative_base

from covalent._shared_files.util_classes import RESULT_STATUS

Base = declarative_base()


class Lattice(Base):
    __tablename__ = "lattices"
    __table_args__ = (UniqueConstraint("dispatch_id", name="u_dispatch_id"),)

    id = Column(Integer, primary_key=True)
    dispatch_id = Column(String(64), nullable=False)

    # id of node if the lattice is actually a sublattice
    electron_id = Column(Integer)

    # name of the lattice function
    name = Column(Text, nullable=False)

    # name of the file containing the lattice function's docstring
    docstring_filename = Column(Text)

    # Workflow status
    status = Column(String(24), nullable=False)

    # Number of nodes in the lattice
    electron_num = Column(Integer, nullable=False)

    # Number of completed nodes in the lattice
    completed_electron_num = Column(Integer, nullable=False)

    # Storage backend type for data files ("local", "s3")
    storage_type = Column(Text)

    # Bucket name (dispatch_id)
    storage_path = Column(Text)

    # Name of the file containing the serialized function
    function_filename = Column(Text)

    # Name of the file containing the function string
    function_string_filename = Column(Text)

    # Short name describing the executor ("local", "dask", etc)
    executor = Column(Text)

    # JSONified executor attributes
    executor_data = Column(Text)

    # Short name describing the workflow executor ("local", "dask", etc)
    workflow_executor = Column(Text)

    # JSONified executor attributes
    workflow_executor_data = Column(Text)

    # Name of the file containing an error message for the workflow
    error_filename = Column(Text)

    # Name of the file containing the serialized input data
    inputs_filename = Column(Text)

    # Name of the file containing the serialized named args
    named_args_filename = Column(Text)

    # Name of the file containing the serialized named kwargs
    named_kwargs_filename = Column(Text)

    # name of the file containing the serialized output
    results_filename = Column(Text)

    # Name of the file containing the default electron hooks
    hooks_filename = Column(Text)

    # Name of the file containing the set of cova imports
    cova_imports_filename = Column(Text)

    # Name of the file containing the set of lattice imports
    lattice_imports_filename = Column(Text)

    # Results directory (will be deprecated soon)
    results_dir = Column(Text)

    # Dispatch id of the root lattice in a hierarchy of sublattices
    root_dispatch_id = Column(String(64), nullable=True)

    # Name of the column which signifies soft deletion of a lattice
    is_active = Column(Boolean, nullable=False, default=True)

    # Python version
    python_version = Column(Text)

    # Covalent SDK version
    covalent_version = Column(Text)

    # Timestamps
    created_at = Column(DateTime, nullable=False, server_default=func.now())
    updated_at = Column(DateTime, nullable=False, onupdate=func.now(), server_default=func.now())
    started_at = Column(DateTime)
    completed_at = Column(DateTime)


class Electron(Base):
    __tablename__ = "electrons"
    __table_args__ = (Index("latid_nodeid_idx", "parent_lattice_id", "transport_graph_node_id"),)
    id = Column(Integer, primary_key=True)

    # id of the lattice containing this electron
    parent_lattice_id = Column(Integer, ForeignKey("lattices.id"), nullable=False)

    # id of the node in the context of a transport graph
    transport_graph_node_id = Column(Integer, nullable=False)

    # id of the node's task group in the context of a transport graph
    task_group_id = Column(Integer, nullable=False)

    # Node type
    type = Column(String(24), nullable=False)

    # Node name
    name = Column(Text, nullable=False)

    # Execution status of the node
    status = Column(String(24), nullable=False)

    # Storage backend type for data files ("local", "s3")
    storage_type = Column(Text)

    # Bucket name (dispatch_id)
    storage_path = Column(Text)

    # Name of the file containing the serialized function
    function_filename = Column(Text)

    # Name of the file containing the function string
    function_string_filename = Column(Text)

    # Short name describing the executor ("local", "dask", etc)
    executor = Column(Text)

    # JSONified executor attributes
    executor_data = Column(Text)

    # name of the file containing the serialized output
    results_filename = Column(Text)

    # Name of the file containing the serialized parameter (for parameter nodes)
    value_filename = Column(Text)

    # Name of the file containing standard output generated by the task
    stdout_filename = Column(Text)

    # Name of the file containing the electron execution dependencies
    hooks_filename = Column(Text)

    # Whether qelectron data exists or not
    qelectron_data_exists = Column(Boolean, nullable=False, default=False)

    # Name of the file containing standard error generated by the task
    stderr_filename = Column(Text)

    # Name of the file containing errors generated by the task runner or executor
    error_filename = Column(Text)

    # Name of the column which signifies soft deletion of the electrons corresponding to a lattice
    is_active = Column(Boolean, nullable=False, default=True)

    # Foreign key reference to Jobs table
    job_id = Column(Integer, ForeignKey("jobs.id", name="job_id_link"), nullable=False)

    # Timestamps
    created_at = Column(DateTime, nullable=False, server_default=func.now())
    updated_at = Column(DateTime, nullable=False, onupdate=func.now(), server_default=func.now())
    started_at = Column(DateTime)
    completed_at = Column(DateTime)


class ElectronDependency(Base):
    __tablename__ = "electron_dependency"
    id = Column(Integer, primary_key=True)
    __table_args__ = (Index("cnode_idx", "electron_id"), Index("pnode_idx", "parent_electron_id"))
    # Unique ID of electron
    electron_id = Column(
        Integer, ForeignKey("electrons.id", name="child_electron_link"), nullable=False
    )

    # Unique ID of the electron's parent
    parent_electron_id = Column(
        Integer, ForeignKey("electrons.id", name="parent_electron_link"), nullable=False
    )

    edge_name = Column(Text, nullable=False)

    # arg, kwarg, null
    parameter_type = Column(String(24), nullable=True)

    # Argument position - this value is null unless parameter_type is arg
    arg_index = Column(Integer, nullable=True)

    # Name of the column which signifies soft deletion of the electron dependencies corresponding to a lattice
    is_active = Column(Boolean, nullable=False, default=True)

    updated_at = Column(DateTime, nullable=True, onupdate=func.now(), server_default=func.now())
    created_at = Column(DateTime, nullable=False, server_default=func.now())


class Job(Base):
    __tablename__ = "jobs"
    id = Column(Integer, primary_key=True)

    # Indicates whether the job has been requested to be cancelled
    cancel_requested = Column(Boolean, nullable=False, default=False)

    # Job state -- to be filtered/interpreted by each plugin
    status = Column(String(24), nullable=False, default=str(RESULT_STATUS.NEW_OBJECT))

    # JSON-serialized identifier for job
    job_handle = Column(Text, nullable=False, default="null")


# Core Lattice assets
class LatticeAsset(Base):
    __tablename__ = "lattice_assets"
    __table_args__ = (Index("lattice_assets_idx", "meta_id", "key"),)

    id = Column(Integer, primary_key=True)

    # Lattice record id
    meta_id = Column(Integer, ForeignKey("lattices.id", name="lattice_link"), nullable=False)

    # Asset record id
    asset_id = Column(Integer, ForeignKey("assets.id", name="asset_link"), nullable=False)

    # Asset key
    key = Column(String(24), nullable=False)


# Core Electron assets
class ElectronAsset(Base):
    __tablename__ = "electron_assets"
    __table_args__ = (Index("electron_assets_idx", "meta_id", "key"),)
    id = Column(Integer, primary_key=True)

    # Electron record id
    meta_id = Column(Integer, ForeignKey("electrons.id", name="electron_link"), nullable=False)

    # Asset record id
    asset_id = Column(Integer, ForeignKey("assets.id", name="asset_link"), nullable=False)

    # Asset key
    key = Column(String(24), nullable=False)


class Asset(Base):
    __tablename__ = "assets"
    id = Column(Integer, primary_key=True)

    # Storage backend type for data files ("local", "s3")
    storage_type = Column(Text, nullable=False)

    # Bucket name
    storage_path = Column(Text, nullable=False)

    # Object key
    object_key = Column(Text, nullable=False)

    # Digest algorithm ("md5", "sha1")
    digest_alg = Column(Text, nullable=True)

    # Hex repr of digest
    digest = Column(Text, nullable=True)

    # Remote location of asset
    remote_uri = Column(Text, nullable=True)

    # Size in bytes
    size = Column(Integer, nullable=True)



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import importlib
from functools import lru_cache
from typing import Any, Dict, List, Optional, Union

import orjson
from pydantic import BaseModel

from ..._shared_files.qinfo import QNodeSpecs
from ...executor.qbase import BaseQExecutor

BATCH_ID_SEPARATOR = "@"
MAX_DIFFERENT_EXECUTORS = 10


class CircuitInfo(BaseModel):
    electron_node_id: Optional[int] = None
    dispatch_id: Optional[str] = None
    circuit_name: Optional[str] = None
    circuit_description: Optional[str] = None
    circuit_diagram: Optional[str] = None
    qnode_specs: Optional[Union[Dict[str, Any], QNodeSpecs]] = None
    qexecutor: Optional[BaseQExecutor] = None
    save_time: datetime.datetime
    circuit_id: Optional[str] = None
    qscript: Optional[str] = None
    execution_time: Optional[float] = None
    result: Optional[List[Any]] = None
    result_metadata: Optional[List[Dict[str, Any]]] = None


@lru_cache
def get_cached_module():
    return importlib.import_module(".executor", package="covalent")


def executor_from_dict(executor_dict: Dict):
    if "executors" in executor_dict:
        executors = [executor_from_dict(ed) for ed in executor_dict["executors"]]
        executor_dict["executors"] = executors

    name = executor_dict["name"]
    executor_class = getattr(get_cached_module(), name)
    return executor_class(**executor_dict)


@lru_cache(maxsize=MAX_DIFFERENT_EXECUTORS)
def get_cached_executor(**executor_dict):
    if "executors" in executor_dict:
        executors = tuple(orjson.loads(ex) for ex in executor_dict["executors"])
        executor_dict["executors"] = executors

    return executor_from_dict(executor_dict)


def reconstruct_executors(deconstructed_executors: List[Dict]):
    return [executor_from_dict(de) for de in deconstructed_executors]


def get_circuit_id(batch_id, circuit_number):
    return f"circuit_{circuit_number}{BATCH_ID_SEPARATOR}{batch_id}"



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""API client"""

import json
import os
from typing import Dict

import requests
from requests.adapters import HTTPAdapter


class CovalentAPIClient:
    """Thin wrapper around Requests to centralize error handling."""

    def __init__(self, dispatcher_addr: str, adapter: HTTPAdapter = None, auto_raise: bool = True):
        self.dispatcher_addr = dispatcher_addr
        self.adapter = adapter
        self.auto_raise = auto_raise

    def prepare_headers(self, **kwargs):
        extra_headers = CovalentAPIClient.get_extra_headers()
        headers = kwargs.get("headers", {})
        if headers:
            kwargs.pop("headers")
        headers.update(extra_headers)
        return headers

    def get(self, endpoint: str, **kwargs):
        headers = self.prepare_headers(**kwargs)
        url = self.dispatcher_addr + endpoint
        try:
            with requests.Session() as session:
                if self.adapter:
                    session.mount("http://", self.adapter)

                r = session.get(url, headers=headers, **kwargs)

            if self.auto_raise:
                r.raise_for_status()

        except requests.exceptions.ConnectionError:
            message = f"The Covalent server cannot be reached at {url}. Local servers can be started using `covalent start` in the terminal. If you are using a remote Covalent server, contact your systems administrator to report an outage."
            print(message)
            raise

        return r

    def put(self, endpoint: str, **kwargs):
        headers = self.prepare_headers()
        url = self.dispatcher_addr + endpoint
        try:
            with requests.Session() as session:
                if self.adapter:
                    session.mount("http://", self.adapter)

                r = session.put(url, headers=headers, **kwargs)

            if self.auto_raise:
                r.raise_for_status()
        except requests.exceptions.ConnectionError:
            message = f"The Covalent server cannot be reached at {url}. Local servers can be started using `covalent start` in the terminal. If you are using a remote Covalent server, contact your systems administrator to report an outage."
            print(message)
            raise

        return r

    def post(self, endpoint: str, **kwargs):
        headers = self.prepare_headers()
        url = self.dispatcher_addr + endpoint
        try:
            with requests.Session() as session:
                if self.adapter:
                    session.mount("http://", self.adapter)

                r = session.post(url, headers=headers, **kwargs)

            if self.auto_raise:
                r.raise_for_status()
        except requests.exceptions.ConnectionError:
            message = f"The Covalent server cannot be reached at {url}. Local servers can be started using `covalent start` in the terminal. If you are using a remote Covalent server, contact your systems administrator to report an outage."
            print(message)
            raise

        return r

    def delete(self, endpoint: str, **kwargs):
        headers = self.prepare_headers()
        url = self.dispatcher_addr + endpoint
        try:
            with requests.Session() as session:
                if self.adapter:
                    session.mount("http://", self.adapter)

                r = session.delete(url, headers=headers, **kwargs)

            if self.auto_raise:
                r.raise_for_status()
        except requests.exceptions.ConnectionError:
            message = f"The Covalent server cannot be reached at {url}. Local servers can be started using `covalent start` in the terminal. If you are using a remote Covalent server, contact your systems administrator to report an outage."
            print(message)
            raise

        return r

    @classmethod
    def get_extra_headers(headers: Dict) -> Dict:
        # This is expected to be a JSONified dictionary
        data = os.environ.get("COVALENT_EXTRA_HEADERS")
        if data:
            return json.loads(data)
        else:
            return {}



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from covalent._shared_files.config import get_config

log_level = get_config("sdk.log_level").upper()
log_to_file = get_config("sdk.enable_logging").upper() == "TRUE"


def log_config():
    """
    Get Log configurations
    Return:
        Log configuration with respect to settings
    """
    log = log_level if log_to_file else None

    return {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "default": {
                "()": "uvicorn.logging.DefaultFormatter",
                "datefmt": "%Y-%m-%d %H:%M:%S",
                "format": "[%(asctime)s,%(msecs)03d] [%(levelname)s] %(message)s",
            },
            "access": {
                "()": "uvicorn.logging.AccessFormatter",
                "datefmt": "%Y-%m-%d %H:%M:%S",
                "format": "[%(asctime)s,%(msecs)03d] [%(levelname)s] %(message)s",
            },
        },
        "handlers": {
            "default": {
                "formatter": "default",
                "class": "logging.StreamHandler",
                "stream": "ext://sys.stderr",
            },
            "access": {
                "formatter": "access",
                "class": "logging.StreamHandler",
                "stream": "ext://sys.stdout",
            },
        },
        "loggers": {
            "uvicorn.error": {"level": log, "handlers": ["default"], "propogate": False},
            "uvicorn.access": {"level": log, "handlers": ["access"], "propagate": "no"},
        },
    }



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations

from typing import Generic, Type, TypeVar

from sqlalchemy import select, update
from sqlalchemy.orm import Session, load_only

from .._db import models

T = TypeVar("T", bound=models.Base)


class Record(Generic[T]):
    """
    Thin wrapper for a SQLALchemy record
    """

    @classmethod
    @property
    def model(cls) -> Type[T]:
        raise NotImplementedError

    def __init__(self, session: Session, record: models.Base, *, fields: list):
        self._id = record.id
        self._attrs = {k: getattr(record, k) for k in fields}

    @property
    def primary_key(self):
        return self._id

    @classmethod
    def get(
        cls,
        session: Session,
        *,
        fields: list,
        equality_filters: dict,
        membership_filters: dict,
        for_update: bool = False,
    ):
        """Bulk ORM-enabled SELECT.

        Args:
            session: SQLalchemy session
            fields: List of columns to select
            equality_filters: Dict{field_name: value}
            membership_filters: Dict{field_name: value_list}
            for_update: Whether to lock the selected rows

        """
        stmt = select(cls.model)
        for attr, val in equality_filters.items():
            stmt = stmt.where(getattr(cls.model, attr) == val)
        for attr, vals in membership_filters.items():
            stmt = stmt.where(getattr(cls.model, attr).in_(vals))
        if len(fields) > 0:
            attrs = [getattr(cls.model, f) for f in fields]
            stmt = stmt.options(load_only(*attrs))
        if for_update:
            stmt = stmt.with_for_update()

        return session.scalars(stmt).all()

    @classmethod
    def get_by_primary_key(
        cls, session: Session, primary_key: int, *, for_update: bool = False
    ) -> T:
        return session.get(cls.model, primary_key, with_for_update=for_update)

    @classmethod
    def create(cls, session: Session, *, insert_kwargs: dict, flush: bool = True) -> T:
        """Create a new record.

        Args:
            session: SQLalchemy session
            insert_kwargs: kwargs to pass to the model constructor
            flush: Whether to flush the session immediately

        Returns: A SQLAlchemy model of type T. If `flush=False`, the
            model will need to be added to the session manually.

        """

        new_record = cls.model(**insert_kwargs)
        session.add(new_record)
        if flush:
            session.flush()
        return new_record

    @classmethod
    def update_bulk(
        cls, session: Session, *, values: dict, equality_filters: dict, membership_filters: dict
    ):
        """Bulk update.

        Args:
            session: SQLAlchemy session
            values: dictionary of values to pass to UPDATE
            equality_filters: Dict{field_name: value}
            membership_filters: Dict{field_name: value_list}
        """

        stmt = update(cls.model).values(**values)
        for attr, val in equality_filters.items():
            stmt = stmt.where(getattr(cls.model, attr) == val)
        for attr, vals in membership_filters.items():
            stmt = stmt.where(getattr(cls.model, attr).in_(vals))
        session.execute(stmt)

    @classmethod
    def incr_bulk(
        cls,
        session: Session,
        *,
        increments: dict,
        equality_filters: dict,
        membership_filters: dict,
    ):
        """Bulk increment numerical fields

        Args:
            session: SQLAlchemy session
            increment: dictionary {field: delta}
            equality_filters: Dict{field_name: value}
            membership_filters: Dict{field_name: value_list}
        """

        kwargs = {}
        for field, delta in increments.items():
            col = getattr(cls.model, field)
            kwargs[field] = col + delta

        stmt = update(cls.model).values(**kwargs)
        for attr, val in equality_filters.items():
            stmt = stmt.where(getattr(cls.model, attr) == val)
        for attr, vals in membership_filters.items():
            stmt = stmt.where(getattr(cls.model, attr).in_(vals))
        session.execute(stmt)

    def update(self, session: Session, *, values: dict):
        """Update the corresponding DB record."""

        type(self).update_bulk(
            session,
            values=values,
            equality_filters={"id": self.primary_key},
            membership_filters={},
        )
        for k, v in values.items():
            self._attrs[k] = v

    def incr(self, session: Session, *, increments: dict):
        """Increment the fields of the corresponding record."""
        type(self).incr_bulk(
            session,
            increments=increments,
            equality_filters={"id": self.primary_key},
            membership_filters={},
        )

    def refresh(self, session: Session, *, fields: set, for_update: bool = False):
        """Sync with DB"""
        records = type(self).get(
            session,
            fields=fields,
            equality_filters={"id": self._id},
            membership_filters={},
            for_update=for_update,
        )
        record = records[0]
        self._attrs = {k: getattr(record, k) for k in fields}

    @property
    def attrs(self) -> dict:
        return self._attrs

    def __contains__(self, item: str):
        return item in self._attrs



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from threading import Event

from .base import BaseTrigger


class TimeTrigger(BaseTrigger):
    """
    Performs a trigger action every `time_gap` seconds.

    Args:
        time_gap: Amount of seconds to wait before doing a trigger action

    Attributes:
        self.time_gap: Amount of seconds to wait before doing a trigger action
        self.stop_flag: Thread safe flag used to check whether the stop condition has been met
    """

    def __init__(
        self,
        time_gap: int,
        lattice_dispatch_id: str = None,
        dispatcher_addr: str = None,
        triggers_server_addr: str = None,
    ):
        super().__init__(lattice_dispatch_id, dispatcher_addr, triggers_server_addr)
        self.time_gap = time_gap
        self.stop_flag = None

    def observe(self) -> None:
        """
        Keep performing the trigger action every `self.time_gap` seconds
        until stop condition has been met.
        """

        # Stopping mechanism for a blocking observe()
        self.stop_flag = Event()
        while not self.stop_flag.is_set():
            time.sleep(self.time_gap)
            self.trigger()

    def stop(self) -> None:
        """
        Stop the running `self.observe()` method by setting the `self.stop_flag` flag.
        """

        self.stop_flag.set()



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utilities for Qiskit-based QElectron executors and devices
"""

from qiskit_ibm_runtime import Options


def extract_options(opts: dict) -> Options:
    """
    Construct a Qiskit `Options` object from the options dictionary
    """
    if isinstance(opts, tuple):
        opts = dict(opts)

    _options = Options()
    _options.optimization_level = opts.get("optimization_level", 3)
    _options.resilience_level = opts.get("resilience_level", 1)
    _options.max_execution_time = opts.get("max_execution_time", None)
    _options.transpilation = opts.get("transpilation", _options.transpilation)
    _options.resilience = opts.get("resilience", _options.resilience)
    _options.execution = opts.get("execution", _options.execution)
    _options.environment = opts.get("environment", _options.environment)
    _options.simulator = opts.get("simulator", _options.simulator)
    return _options



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Defines the core functionality of the legacy runner
"""

import asyncio
import traceback
from datetime import datetime, timezone
from functools import partial
from typing import Any, Dict, List, Tuple

from covalent._shared_files import logger
from covalent._shared_files.config import get_config
from covalent._shared_files.util_classes import RESULT_STATUS
from covalent._workflow import DepsBash, DepsCall, DepsPip
from covalent.executor.utils.wrappers import wrapper_fn

from . import data_manager as datasvc
from .runner_modules import executor_proxy
from .runner_modules.utils import get_executor

app_log = logger.app_log
log_stack_info = logger.log_stack_info
debug_mode = get_config("sdk.log_level") == "debug"


# Domain: runner
# to be called by _run_abstract_task
async def _get_task_input_values(dispatch_id: str, abs_task_inputs: dict) -> dict:
    node_values = {}
    args = abs_task_inputs["args"]
    for node_id in args:
        value = (await datasvc.electron.get(dispatch_id, node_id, ["output"]))["output"]
        node_values[node_id] = value

    kwargs = abs_task_inputs["kwargs"]
    for key, node_id in kwargs.items():
        value = (await datasvc.electron.get(dispatch_id, node_id, ["output"]))["output"]
        node_values[node_id] = value

    return node_values


# Domain: runner
async def _run_abstract_task(
    dispatch_id: str,
    node_id: int,
    node_name: str,
    abstract_inputs: Dict,
    selected_executor: Any,
) -> None:
    # Resolve abstract task and inputs to their concrete (serialized) values
    timestamp = datetime.now(timezone.utc)

    try:
        serialized_callable = (await datasvc.electron.get(dispatch_id, node_id, ["function"]))[
            "function"
        ]

        input_values = await _get_task_input_values(dispatch_id, abstract_inputs)

        abstract_args = abstract_inputs["args"]
        abstract_kwargs = abstract_inputs["kwargs"]
        args = [input_values[node_id] for node_id in abstract_args]
        kwargs = {k: input_values[v] for k, v in abstract_kwargs.items()}

        task_input = {"args": args, "kwargs": kwargs}

        app_log.debug(f"Collecting deps for task {node_id}")

        call_before, call_after = await _gather_deps(dispatch_id, node_id)

    except Exception as ex:
        app_log.error(f"Exception when trying to resolve inputs or deps: {ex}")
        node_result = datasvc.generate_node_result(
            node_id=node_id,
            start_time=timestamp,
            end_time=timestamp,
            status=RESULT_STATUS.FAILED,
            error=str(ex),
        )
        return node_result

    node_result = datasvc.generate_node_result(
        node_name=node_name,
        node_id=node_id,
        start_time=timestamp,
        status=RESULT_STATUS.RUNNING,
    )
    app_log.debug(f"7: Marking node {node_id} as running (_run_abstract_task)")

    await datasvc.update_node_result(dispatch_id, node_result)

    return await _run_task(
        dispatch_id=dispatch_id,
        node_id=node_id,
        serialized_callable=serialized_callable,
        selected_executor=selected_executor,
        node_name=node_name,
        call_before=call_before,
        call_after=call_after,
        inputs=task_input,
    )


# Domain: runner
async def _run_task(
    dispatch_id: str,
    node_id: int,
    inputs: Dict,
    serialized_callable: Any,
    selected_executor: Any,
    call_before: List,
    call_after: List,
    node_name: str,
) -> None:
    """
    Run a task with given inputs on the selected executor.
    Also updates the status of current node execution while
    checking if a redispatch has occurred. Exclude those nodes
    from execution which were completed.

    Also verifies if execution of this dispatch has been cancelled.

    Args:
        inputs: Inputs for the task.
        node_id: Node id of the task to be executed.

    Returns:
        None
    """

    dispatch_info = await datasvc.dispatch.get(dispatch_id, ["results_dir"])
    results_dir = dispatch_info["results_dir"]

    # Instantiate the executor from JSON
    try:
        app_log.debug(f"Instantiating executor for {dispatch_id}:{node_id}")
        executor = get_executor(
            node_id=node_id,
            selected_executor=selected_executor,
            loop=asyncio.get_running_loop(),
            pool=None,
        )
    except Exception as ex:
        tb = "".join(traceback.TracebackException.from_exception(ex).format())
        app_log.debug("Exception when trying to instantiate executor:")
        app_log.debug(tb)
        error_msg = tb if debug_mode else str(ex)
        node_result = datasvc.generate_node_result(
            node_id=node_id,
            end_time=datetime.now(timezone.utc),
            status=RESULT_STATUS.FAILED,
            error=error_msg,
        )
        return node_result

    # run the task on the executor and register any failures
    try:
        app_log.debug(f"Executing task {node_name}")

        assembled_callable = partial(wrapper_fn, serialized_callable, call_before, call_after)

        # Start listening for messages from the plugin
        asyncio.create_task(executor_proxy.watch(dispatch_id, node_id, executor))

        output, stdout, stderr, status = await executor._execute(
            function=assembled_callable,
            args=inputs["args"],
            kwargs=inputs["kwargs"],
            dispatch_id=dispatch_id,
            results_dir=results_dir,
            node_id=node_id,
        )

        node_result = datasvc.generate_node_result(
            node_id=node_id,
            node_name=node_name,
            end_time=datetime.now(timezone.utc),
            status=status,
            output=output,
            stdout=stdout,
            stderr=stderr,
        )

    except Exception as ex:
        tb = "".join(traceback.TracebackException.from_exception(ex).format())
        app_log.debug(f"Exception occurred when running task {node_id}:")
        app_log.debug(tb)
        error_msg = tb if debug_mode else str(ex)
        node_result = datasvc.generate_node_result(
            node_id=node_id,
            node_name=node_name,
            end_time=datetime.now(timezone.utc),
            status=RESULT_STATUS.FAILED,
            error=error_msg,
        )
    app_log.debug(f"Node result: {node_result}")
    return node_result


# Domain: runner
async def _gather_deps(dispatch_id: str, node_id: int) -> Tuple[List, List]:
    """Assemble deps for a node into the final call_before and call_after"""

    resp = await datasvc.electron.get(dispatch_id, node_id, ["hooks"])

    hooks = resp["hooks"]

    # Assemble call_before and call_after from all the deps
    deps = hooks.get("deps", {})
    call_before_objs_json = hooks.get("call_before", [])
    call_after_objs_json = hooks.get("call_after", [])

    call_before = []
    call_after = []

    # Rehydrate deps from JSON
    if "bash" in deps:
        dep = DepsBash()
        dep.from_dict(deps["bash"])
        call_before.append(dep.apply())

    if "pip" in deps:
        dep = DepsPip()
        dep.from_dict(deps["pip"])
        call_before.append(dep.apply())

    for dep_json in call_before_objs_json:
        dep = DepsCall()
        dep.from_dict(dep_json)
        call_before.append(dep.apply())

    for dep_json in call_after_objs_json:
        dep = DepsCall()
        dep.from_dict(dep_json)
        call_after.append(dep.apply())

    return call_before, call_after


# Domain: runner
async def run_abstract_task(
    dispatch_id: str,
    node_id: int,
    node_name: str,
    abstract_inputs: Dict,
    selected_executor: Any,
) -> None:
    node_result = await _run_abstract_task(
        dispatch_id=dispatch_id,
        node_id=node_id,
        node_name=node_name,
        abstract_inputs=abstract_inputs,
        selected_executor=selected_executor,
    )
    await datasvc.update_node_result(dispatch_id, node_result)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Mappings between result attributes and DB records"""


from covalent._shared_files.schemas import result
from covalent._shared_files.util_classes import Status

from . import lattice_utils

ATTRIBUTES = {
    "start_time",
    "end_time",
    "results_dir",
    "lattice",
    "dispatch_id",
    "root_dispatch_id",
    "electron_id",
    "status",
    "task_failed",
    "task_cancelled",
    "result",
    "num_nodes",
    "error",
}

METADATA_KEYS = result.METADATA_KEYS.copy()
METADATA_KEYS.update({"results_dir", "electron_id"})
ASSET_KEYS = result.ASSET_KEYS


_meta_record_map = {
    "start_time": "started_at",
    "end_time": "completed_at",
    "results_dir": "results_dir",
    "dispatch_id": "dispatch_id",
    "root_dispatch_id": "root_dispatch_id",
    "electron_id": "electron_id",
    "status": "status",
    "num_nodes": "electron_num",
    "completed_electron_num": "completed_electron_num",
}

_db_meta_record_map = {
    "id": "id",
    "electron_id": "electron_id",
    "storage_path": "storage_path",
    "storage_type": "storage_type",
    "completed_electron_num": "completed_electron_num",
}

_meta_record_map.update(_db_meta_record_map)
_meta_record_map.update(lattice_utils._meta_record_map)


# Obsoleted by LatticeAsset table
_asset_record_map = {
    "result": "results_filename",
    "error": "error_filename",
}


def get_status_filter(raw: str):
    return Status(raw)


def set_status_filter(stat: Status):
    return str(stat)


get_filters = {key: lambda x: x for key in METADATA_KEYS.union(ASSET_KEYS)}

set_filters = {key: lambda x: x for key in METADATA_KEYS.union(ASSET_KEYS)}

custom_get_filters = {"status": get_status_filter, "completed_electron_num": lambda x: x}

custom_set_filters = {"status": set_status_filter, "completed_electron_num": lambda x: x}

get_filters.update(custom_get_filters)
set_filters.update(custom_set_filters)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from subprocess import PIPE, CalledProcessError, Popen
from typing import Optional

from .. import File
from .transfer_strategy_base import FileTransferStrategy


class Rsync(FileTransferStrategy):
    """
    Implements Base FileTransferStrategy class to use rsync to move files to and from remote or local filesystems. Rsync via ssh is used if one of the provided files is marked as remote.

    Attributes:
        user: (optional) Determine user to specify for remote host if using rsync with ssh
        host: (optional) Determine what host to connect to if using rsync with ssh
        private_key_path: (optional) Filepath for ssh private key to use if using rsync with ssh
    """

    def __init__(
        self,
        user: Optional[str] = "",
        host: Optional[str] = "",
        private_key_path: Optional[str] = None,
    ):
        self.user = user
        self.private_key_path = private_key_path
        self.host = host

        if self.private_key_path and not os.path.exists(self.private_key_path):
            raise FileNotFoundError(
                f"Provided private key ({self.private_key_path}) does not exist. Could not instantiate Rsync File Transfer Strategy. "
            )

    def get_rsync_ssh_cmd(
        self, local_file: File, remote_file: File, transfer_from_remote: bool = False
    ) -> str:
        local_filepath = local_file.filepath
        remote_filepath = remote_file.filepath
        args = ["rsync"]
        if self.private_key_path:
            args.append(f'-ae "ssh -i {self.private_key_path}"')
        else:
            args.append("-ae ssh")

        remote_source = f"{self.user}@{self.host}:{remote_filepath}"

        if transfer_from_remote:
            args.append(remote_source)
            args.append(local_filepath)
        else:
            args.append(local_filepath)
            args.append(remote_source)

        return " ".join(args)

    def get_rsync_cmd(
        self, from_file: File, to_file: File, transfer_from_remote: bool = False
    ) -> str:
        from_filepath = from_file.filepath
        to_filepath = to_file.filepath
        return f"rsync -a {from_filepath} {to_filepath}"

    def return_subprocess_callable(self, cmd) -> None:
        def callable():
            p = Popen(cmd, shell=True, stdout=PIPE, stderr=PIPE)
            output, error = p.communicate()
            if p.returncode != 0:
                raise CalledProcessError(p.returncode, f'"{cmd}" with error: {str(error)}')

        return callable

    # return callable to move files in the local file system
    def cp(self, from_file: File, to_file: File = File()) -> None:
        cmd = self.get_rsync_cmd(from_file, to_file)
        return self.return_subprocess_callable(cmd)

    # return callable to download here implies 'from' is a remote source
    def download(self, from_file: File, to_file: File = File()) -> File:
        cmd = self.get_rsync_ssh_cmd(to_file, from_file, transfer_from_remote=True)
        return self.return_subprocess_callable(cmd)

    # return callable to upload here implies 'to' is a remote source
    def upload(self, from_file: File, to_file: File) -> None:
        cmd = self.get_rsync_ssh_cmd(from_file, to_file, transfer_from_remote=False)
        return self.return_subprocess_callable(cmd)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base storage backend provider"""

from dataclasses import dataclass
from typing import Optional, Tuple


@dataclass
class Digest:
    algorithm: str
    hexdigest: str


class BaseProvider:
    @classmethod
    @property
    def scheme(cls) -> str:
        raise NotImplementedError

    def digest(self, bucket_name: str, object_key: str) -> Digest:
        raise NotImplementedError

    def size(self, bucket_name: str, object_key: str) -> int:
        raise NotImplementedError

    def get_uri_components(
        self, dispatch_id: str, node_id: Optional[int], asset_key: str
    ) -> Tuple[str, str]:
        """Compute storage_path and object_key for a workflow asset.

        Args:
            dispatch_id: The workflow dispatch id
            node_id: The electron's node id or `None` if the asset has workflow scope.
            asset_key: The key describing the asset.

        Returns:
            storage_path, object_key

        The semantics `storage_path` and `object_key` may differ
        slightly between backends but are constrained by the requirement that
        `{scheme}://{storage_path}/{object_key}` is a valid URI for
        the asset.

        """

        raise NotImplementedError



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Functionality for importing dispatch submissions
"""

import uuid
from typing import Optional

from covalent._shared_files import logger
from covalent._shared_files.config import get_config
from covalent._shared_files.schemas.result import ResultSchema
from covalent_dispatcher._dal.asset import copy_asset
from covalent_dispatcher._dal.importers.result import handle_redispatch, import_result
from covalent_dispatcher._dal.result import Result as SRVResult

from .utils import dm_pool, run_in_executor

BASE_PATH = get_config("dispatcher.results_dir")

app_log = logger.app_log

# Concurrent futures for copying assets during redispatch
copy_futures = {}


# Domain: result
def get_unique_id() -> str:
    """
    Get a unique ID.

    Args:
        None

    Returns:
        str: Unique ID
    """

    return str(uuid.uuid4())


def _import_manifest(
    res: ResultSchema,
    parent_dispatch_id: Optional[str],
    parent_electron_id: Optional[int],
) -> ResultSchema:
    if not res.metadata.dispatch_id:
        res.metadata.dispatch_id = get_unique_id()

    # Compute root_dispatch_id for sublattice dispatches
    if parent_dispatch_id:
        parent_result_object = SRVResult.from_dispatch_id(
            dispatch_id=parent_dispatch_id,
            bare=True,
        )
        res.metadata.root_dispatch_id = parent_result_object.root_dispatch_id
    else:
        res.metadata.root_dispatch_id = res.metadata.dispatch_id

    return import_result(res, BASE_PATH, parent_electron_id)


def _get_all_assets(dispatch_id: str):
    result_object = SRVResult.from_dispatch_id(dispatch_id, bare=True)
    return result_object.get_all_assets()


def _pull_assets(manifest: ResultSchema) -> None:
    dispatch_id = manifest.metadata.dispatch_id
    assets = _get_all_assets(dispatch_id)
    futs = []
    for asset in assets["lattice"]:
        if asset.remote_uri:
            asset.download(asset.remote_uri)

    for asset in assets["nodes"]:
        if asset.remote_uri:
            asset.download(asset.remote_uri)

    app_log.debug(f"imported {len(futs)} assets for dispatch {dispatch_id}")


async def import_manifest(
    manifest: ResultSchema,
    parent_dispatch_id: Optional[str],
    parent_electron_id: Optional[int],
) -> ResultSchema:
    filtered_manifest = await run_in_executor(
        _import_manifest, manifest, parent_dispatch_id, parent_electron_id
    )
    await run_in_executor(_pull_assets, filtered_manifest)

    return filtered_manifest


def _copy_assets(assets_to_copy):
    for item in assets_to_copy:
        src, dest = item
        copy_asset(src, dest)


def _import_derived_manifest(
    manifest: ResultSchema,
    parent_dispatch_id: str,
    reuse_previous_results: bool,
) -> ResultSchema:
    filtered_manifest = _import_manifest(manifest, None, None)
    filtered_manifest, assets_to_copy = handle_redispatch(
        filtered_manifest, parent_dispatch_id, reuse_previous_results
    )

    dispatch_id = filtered_manifest.metadata.dispatch_id
    fut = dm_pool.submit(_copy_assets, assets_to_copy)
    copy_futures[dispatch_id] = fut
    fut.add_done_callback(lambda x: copy_futures.pop(dispatch_id))

    return filtered_manifest


async def import_derived_manifest(
    manifest: ResultSchema,
    parent_dispatch_id: str,
    reuse_previous_results: bool,
) -> ResultSchema:
    filtered_manifest = await run_in_executor(
        _import_derived_manifest,
        manifest,
        parent_dispatch_id,
        reuse_previous_results,
    )

    await run_in_executor(_pull_assets, filtered_manifest)

    return filtered_manifest



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from logging.config import fileConfig

from alembic import context
from sqlalchemy import engine_from_config, pool

from covalent_dispatcher._db.datastore import DataStore
from covalent_dispatcher._db.models import Base

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

# Interpret the config file for Python logging.
# This line sets up loggers basically.
if config.config_file_name is not None and config.attributes.get("configure_logger", True):
    fileConfig(config.config_file_name)

target_metadata = Base.metadata

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.


def run_migrations_offline() -> None:
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    db = DataStore.factory()
    url = db.db_URL
    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
        render_as_batch=True,
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online() -> None:
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    configuration = config.get_section(config.config_ini_section)
    db = DataStore.factory()
    configuration["sqlalchemy.url"] = db.db_URL
    connectable = engine_from_config(
        configuration,
        prefix="sqlalchemy.",
        poolclass=pool.NullPool,
    )

    with connectable.connect() as connection:
        context.configure(
            connection=connection, target_metadata=target_metadata, render_as_batch=True
        )

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path

from ..._shared_files.config import get_config
from .serialize import JsonLmdb, Strategy
from .utils import CircuitInfo


def set_serialization_strategy(strategy_name):
    """
    Select a serialization strategy for the database
    """
    Database.serialization_strategy = strategy_name


class Database:
    # TODO: Should probably revisit the path mechanism later as it
    #       can be implemented in a more elegant way.

    # dash-separated names result in fallback strategy with try/except loops,
    # other valid strategy names uses the one strategy every time
    # see `covalent_qelectron/quantum_server/serialize.py`
    serialization_strategy = (Strategy.PICKLE, Strategy.ORJSON)

    @property
    def strategy_name(self):
        # using a property here for dynamic access
        # allows runtime strategy selection with `set_serialization_strategy()`
        return Database.serialization_strategy

    def __init__(self, db_dir=None):
        if db_dir:
            self.db_dir = Path(db_dir)
        else:
            self.db_dir = Path(get_config("dispatcher")["qelectron_db_path"])

    def get_db_path(self, dispatch_id=None, node_id=None, *, mkdir=False, direct_path=False):
        if direct_path:
            # If the .mdb file is directly located in the db_dir
            return self.db_dir.resolve().absolute()

        dispatch_id = "default-dispatch" if dispatch_id is None else dispatch_id
        node_id = "default-node" if node_id is None else node_id
        db_path = self.db_dir.joinpath(dispatch_id, f"node-{node_id}")
        if mkdir:
            db_path.mkdir(parents=True, exist_ok=True)

        return db_path.resolve().absolute()

    def _open(self, dispatch_id, node_id, mkdir=False, direct_path=False):
        db_path = self.get_db_path(dispatch_id, node_id, mkdir=mkdir, direct_path=direct_path)

        if not db_path.exists():
            raise FileNotFoundError(f"Missing database directory {db_path}.")

        return JsonLmdb.open_with_strategy(
            file=str(db_path), flag="c", strategy_name=self.strategy_name
        )

    def set(self, keys, values, *, dispatch_id, node_id, direct_path=False):
        with self._open(dispatch_id, node_id, mkdir=True, direct_path=direct_path) as db:
            for i, circuit_id in enumerate(keys):
                stored_val: dict = db.get(circuit_id, None)
                if stored_val is None:
                    continue

                stored_val.update(values[i])
                values[i] = stored_val

            db.update(dict(zip(keys, values)))

    def get_circuit_ids(self, *, dispatch_id, node_id, direct_path=False):
        with self._open(dispatch_id, node_id, direct_path=direct_path) as db:
            return list(db.keys())

    def get_circuit_info(self, circuit_id, *, dispatch_id, node_id, direct_path=False):
        with self._open(dispatch_id, node_id, direct_path=direct_path) as db:
            return CircuitInfo(**db.get(circuit_id, None))

    def get_db_dict(self, *, dispatch_id, node_id, direct_path=False):
        db_copy = {}
        with self._open(dispatch_id, node_id, direct_path=direct_path) as db:
            for key, value in db.items():
                db_copy[key] = value

        return db_copy



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from contextlib import contextmanager

from pydantic import BaseModel


class Context(BaseModel):
    dispatch_id: str = None
    task_id: int = None


def get_context():
    return current_context


@contextmanager
def set_context(dispatch_id: int, task_id: str):
    global current_context
    global unset_context
    current_context = Context(dispatch_id=dispatch_id, task_id=task_id)
    yield
    current_context = unset_context


unset_context = Context()
current_context = unset_context



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Mappings between lattice attributes and DB records"""

import json

from covalent._shared_files.schemas import lattice

ATTRIBUTES = {
    "workflow_function",
    "workflow_function_string",
    "transport_graph",
    "metadata",
    "name",
    "doc",
    "inputs",
    "named_args",
    "named_kwargs",
    "cova_imports",
    "lattice_imports",
}

METADATA_KEYS = lattice.LATTICE_METADATA_KEYS.copy()
METADATA_KEYS.remove("__name__")
METADATA_KEYS.add("name")

ASSET_KEYS = lattice.LATTICE_ASSET_KEYS.copy()
ASSET_KEYS.remove("__doc__")
ASSET_KEYS.add("doc")


_meta_record_map = {
    "name": "name",
    "python_version": "python_version",
    "covalent_version": "covalent_version",
    "executor": "executor",
    "executor_data": "executor_data",
    "workflow_executor": "workflow_executor",
    "workflow_executor_data": "workflow_executor_data",
}

_db_meta_record_map = {
    "electron_id": "electron_id",
    "id": "id",
    "storage_path": "storage_path",
    "storage_type": "storage_type",
}

_meta_record_map.update(_db_meta_record_map)

# Obsoleted by LatticeAsset table
_asset_record_map = {
    "workflow_function": "function_filename",
    "workflow_function_string": "function_string_filename",
    "doc": "docstring_filename",
    "inputs": "inputs_filename",
    "named_args": "named_args_filename",
    "named_kwargs": "named_kwargs_filename",
    "cova_imports": "cova_imports_filename",
    "lattice_imports": "lattice_imports_filename",
    "executor_data": "executor_data_filename",
    "workflow_executor_data": "workflow_executor_data_filename",
    "hooks": "hooks_filename",
}


def get_executor_data_filter(raw: str):
    return json.loads(raw)


def set_executor_data_filter(object_dict: dict):
    return json.dumps(object_dict)


def identity(x):
    return x


custom_get_filters = {
    "executor_data": get_executor_data_filter,
    "workflow_executor_data": get_executor_data_filter,
}
custom_set_filters = {
    "executor_data": set_executor_data_filter,
    "workflow_executor_data": set_executor_data_filter,
}


get_filters = {key: identity for key in METADATA_KEYS.union(ASSET_KEYS)}

set_filters = {key: identity for key in METADATA_KEYS.union(ASSET_KEYS)}

get_filters.update(custom_get_filters)
set_filters.update(custom_set_filters)
